<?php

namespace Drupal\sidekick;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityRepository;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Image\ImageFactory;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\file\Entity\File;
use Drupal\file\Plugin\Field\FieldWidget\FileWidget;
use Drupal\image\Entity\ImageStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Extend the FileWidget plugin.
 */
class ImageWidget extends FileWidget {

  /**
   * The ImageFactory service.
   *
   * @var \Drupal\Core\Image\ImageFactory
   */
  protected ImageFactory $imageFactory;

  /**
   * The Element info.
   *
   * @var \Drupal\Core\Render\ElementInfoManagerInterface
   */
  protected ElementInfoManagerInterface $elementInfo;

  /**
   * The Sidekick service.
   *
   * @var \Drupal\sidekick\SidekickService
   */
  protected SidekickService $sidekickService;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\Renderer
   */
  protected Renderer $renderer;

  /**
   * The entity repository.
   *
   * @var \Drupal\Core\Entity\EntityRepository
   */
  protected EntityRepository $entityRepository;

  /**
   * EntityRepository.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected ConfigFactoryInterface $configFactory;

  /**
   * ImageWidget constructor.
   *
   * @param string $plugin_id
   *   The ID of the plugin this definition is being used for.
   * @param array $plugin_definition
   *   An array of plugin information with multiple keys.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the widget is associated.
   * @param array $settings
   *   The widget settings.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\Core\Render\ElementInfoManagerInterface|null $element_info
   *   The element info manager.
   * @param \Drupal\Core\Image\ImageFactory|null $image_factory
   *   The image factory.
   * @param \Drupal\sidekick\SidekickService|null $sidekick_service
   *   The Sidekick service.
   * @param \Drupal\Core\Render\Renderer|null $renderer
   *   The renderer service.
   * @param \Drupal\Core\Entity\EntityRepository|null $entityRepository
   *   The renderer service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface|null $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The factory for configuration objects.
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    array $third_party_settings,
    ?ElementInfoManagerInterface $element_info = NULL,
    ?ImageFactory $image_factory = NULL,
    ?SidekickService $sidekick_service = NULL,
    ?Renderer $renderer = NULL,
    ?EntityRepository $entityRepository = NULL,
    ?EntityTypeManagerInterface $entityTypeManager = NULL,
    ?ConfigFactoryInterface $configFactory = NULL,
  ) {
    parent::__construct(
      $plugin_id,
      $plugin_definition,
      $field_definition,
      $settings,
      $third_party_settings,
      $element_info
    );
    $this->elementInfo = $element_info;
    $this->imageFactory = $image_factory;
    $this->sidekickService = $sidekick_service;
    $this->renderer = $renderer;
    $this->entityRepository = $entityRepository;
    $this->entityTypeManager = $entityTypeManager;
    $this->configFactory = $configFactory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static($plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('element_info'),
      $container->get('image.factory'),
      $container->get('sidekick.service'),
      $container->get('renderer'),
      $container->get('entity.repository'),
      $container->get('entity_type.manager'),
      $container->get('config.factory')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'progress_indicator' => 'throbber',
      'preview_image_style' => 'thumbnail',
      'sidekick_enabled' => 0,
    ] + parent::defaultSettings();
  }

  /**
   * Form API callback: Processes an image_image field element.
   *
   * Expands the image_image type to include the alt and title fields.
   *
   * This method is assigned as a #process callback in formElement() method.
   */
  public static function process(
    $element,
    FormStateInterface $form_state,
    $form,
  ) {
    $item = $element['#value'];
    $item['fids'] = $element['fids']['#value'];

    $element['#theme'] = 'image_widget';

    // Add the image preview.
    if (!empty($element['#files']) && $element['#preview_image_style']) {
      $file = reset($element['#files']);
      $variables = [
        'style_name' => $element['#preview_image_style'],
        'uri' => $file->getFileUri(),
      ];

      $dimension_key = $variables['uri'] . '.image_preview_dimensions';
      // Determine image dimensions.
      if (isset($element['#value']['width']) && isset($element['#value']['height'])) {
        $variables['width'] = $element['#value']['width'];
        $variables['height'] = $element['#value']['height'];
      }
      elseif ($form_state->has($dimension_key)) {
        $variables += $form_state->get($dimension_key);
      }
      else {
        $variables['width'] = 100;
        $variables['height'] = 100;
      }

      $element['preview'] = [
        '#weight' => -10,
        '#theme' => 'image_style',
        '#width' => $variables['width'],
        '#height' => $variables['height'],
        '#style_name' => $variables['style_name'],
        '#uri' => $variables['uri'],
      ];

      // Store the dimensions in the form so the file doesn't have to be
      // accessed again. This is important for remote files.
      $form_state->set(
        $dimension_key,
        [
          'width' => $variables['width'],
          'height' => $variables['height'],
        ]
      );
    }
    elseif (!empty($element['#default_image'])) {
      $default_image = $element['#default_image'];
      $file = File::load($default_image['fid']);
      if (!empty($file)) {
        $element['preview'] = [
          '#weight' => -10,
          '#theme' => 'image_style',
          '#width' => $default_image['width'],
          '#height' => $default_image['height'],
          '#style_name' => $element['#preview_image_style'],
          '#uri' => $file->getFileUri(),
        ];
      }
    }

    // Add the additional alt and title fields.
    $element['alt'] = [
      '#title' => new TranslatableMarkup('Alternative text'),
      '#type' => 'textfield',
      '#default_value' => $item['alt'] ?? '',
      '#description' => new TranslatableMarkup(
        'Short description of the image used by screen readers and displayed when the image is not loaded. This is important for accessibility.'
      ),
      // @see https://www.drupal.org/node/465106#alt-text
      '#maxlength' => 512,
      '#weight' => -12,
      '#access' => (bool) $item['fids'] && $element['#alt_field'],
      '#required' => $element['#alt_field_required'],
      '#element_validate' => $element['#alt_field_required'] == 1 ? [
        [
          static::class,
          'validateRequiredFields',
        ],
      ] : [],
    ];
    $element['title'] = [
      '#type' => 'textfield',
      '#title' => new TranslatableMarkup('Title'),
      '#default_value' => $item['title'] ?? '',
      '#description' => new TranslatableMarkup(
        'The title is used as a tool tip when the user hovers the mouse over the image.'
      ),
      '#maxlength' => 1024,
      '#weight' => -11,
      '#access' => (bool) $item['fids'] && $element['#title_field'],
      '#required' => $element['#title_field_required'],
      '#element_validate' => $element['#title_field_required'] == 1 ? [
        [
          static::class,
          'validateRequiredFields',
        ],
      ] : [],
    ];

    $element['sidekick'] = [
      '#type' => 'html_tag',
      '#tag' => 'div',
    ];

    $field_name = $element['#field_name'];
    $wrapper_id = 'edit-' . str_replace(
        '_',
        '-',
        strtolower($field_name)
      ) . '-0-alt';
    $element['sidekick'][$field_name . '_sidekick'] = [
      '#type' => 'button',
      '#name' => $wrapper_id . '__sidekick',
      '#value' => '',
      '#executes_submit_callback' => FALSE,
      '#limit_validation_errors' => [
        array_merge(
          $element['#parents'],
          [$field_name]
        ),
      ],
      '#sidekick_module' => 'alt_tag_generator',
      '#access' => (bool) $item['fids'] && $element['#sidekick_enabled'],
      '#field_name' => $field_name,
      '#weight' => -12,
      '#ajax' => [
        'callback' => 'sidekick_image_ajax_callback',
        'wrapper' => $wrapper_id,
      ],
    ];
    $element['sidekick'][$field_name . '_sidekick']['#attributes']['class'][] = 'form-item';
    $element['sidekick'][$field_name . '_sidekick']['#attributes']['class'][] = 'sidekick-button-icon';
    $element['#attached']['library'][] = 'sidekick/sidekick_css';

    return parent::process($element, $form_state, $form);
  }

  /**
   * Validate callback for alt and title field, if the user wants them required.
   *
   * This is separated in a validate function instead of a #required flag to
   * avoid being validated on the process callback.
   */
  public static function validateRequiredFields(
    $element,
    FormStateInterface $form_state,
  ) {
    // Only do validation if the function is triggered from other places than
    // the image process form.
    $triggering_element = $form_state->getTriggeringElement();
    if (!empty($triggering_element['#submit']) && in_array(
        'file_managed_file_submit',
        $triggering_element['#submit'],
        TRUE
      )) {
      $form_state->setLimitValidationErrors([]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = parent::settingsForm($form, $form_state);
    $config = $this->configFactory->get('sidekick.settings');
    $api_key = $config->get('api_key');
    $language_code = $form_state->get('langcode');
    $sidekick_service = $this->sidekickService;
    $sidekickKeyStatus = $sidekick_service->checkKeyStatus(
      $language_code,
      $api_key
    );

    if (!empty($sidekickKeyStatus['status']) && $sidekickKeyStatus['status'] == 200) {
      $element['preview_image_style'] = [
        '#title' => $this->t('Preview image style'),
        '#type' => 'select',
        '#options' => image_style_options(FALSE),
        '#empty_option' => '<' . $this->t('no preview') . '>',
        '#default_value' => $this->getSetting('preview_image_style'),
        '#description' => $this->t(
          'The preview image will be shown while editing the content.'
        ),
        '#weight' => 15,
      ];

      $element['sidekick_enabled'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Enabled Sidekick'),
        '#default_value' => $this->getSetting('sidekick_enabled'),
        '#description' => $this->t('If checked, Enable Sidekick API.'),
      ];
    }

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();

    $image_styles = image_style_options(FALSE);
    // Unset possible 'No defined styles' option.
    unset($image_styles['']);
    // Styles could be lost because of enabled/disabled modules that defines
    // their styles in code.
    $image_style_setting = $this->getSetting('preview_image_style');
    if (isset($image_styles[$image_style_setting])) {
      $preview_image_style = $this->t(
        'Preview image style: @style',
        ['@style' => $image_styles[$image_style_setting]]
      );
    }
    else {
      $preview_image_style = $this->t('No preview');
    }

    array_unshift($summary, $preview_image_style);

    $sidekick_enabled_setting = $this->getSetting('sidekick_enabled');
    if ($sidekick_enabled_setting) {
      array_unshift($summary, $this->t('Sidekick enabled.'));
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(
    FieldItemListInterface $items,
    $delta,
    array $element,
    array &$form,
    FormStateInterface $form_state,
  ) {
    $element = parent::formElement(
      $items,
      $delta,
      $element,
      $form,
      $form_state
    );

    $field_settings = $this->getFieldSettings();

    // Add image validation.
    $element['#upload_validators']['file_validate_is_image'] = [];

    // Add upload resolution validation.
    if ($field_settings['max_resolution'] || $field_settings['min_resolution']) {
      $element['#upload_validators']['file_validate_image_resolution'] = [
        $field_settings['max_resolution'],
        $field_settings['min_resolution'],
      ];
    }

    $extensions = $field_settings['file_extensions'];
    $supported_extensions = $this->imageFactory->getSupportedExtensions();

    // If using custom extension validation, ensure that the extensions are
    // supported by the current image toolkit. Otherwise, validate against all
    // toolkit supported extensions.
    $extensions = !empty($extensions) ? array_intersect(
      explode(' ', $extensions),
      $supported_extensions
    ) : $supported_extensions;
    $element['#upload_validators']['file_validate_extensions'][0] = implode(
      ' ',
      $extensions
    );

    // Add mobile device image capture acceptance.
    $element['#accept'] = 'image/*';

    // Add properties needed by process() method.
    $element['#preview_image_style'] = $this->getSetting(
      'preview_image_style'
    );
    $element['#title_field'] = $field_settings['title_field'];
    $element['#title_field_required'] = $field_settings['title_field_required'];
    $element['#alt_field'] = $field_settings['alt_field'];
    $element['#alt_field_required'] = $field_settings['alt_field_required'];
    $element['#sidekick_enabled'] = $this->getSetting('sidekick_enabled');

    // Default image.
    $default_image = $field_settings['default_image'];
    if (empty($default_image['uuid'])) {
      $default_image = $this->fieldDefinition->getFieldStorageDefinition()
        ->getSetting('default_image');
    }
    // Convert the stored UUID into a file ID.
    if (!empty($default_image['uuid']) && $entity = $this->entityRepository->loadEntityByUuid('file', $default_image['uuid'])) {
      $default_image['fid'] = $entity->id();
    }
    $element['#default_image'] = !empty($default_image['fid']) ? $default_image : [];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    $dependencies = parent::calculateDependencies();
    $style_id = $this->getSetting('preview_image_style');
    /** @var \Drupal\image\ImageStyleInterface $style */
    if ($style_id && $style = ImageStyle::load($style_id)) {
      // If this widget uses a valid image style to display the preview of the
      // uploaded image, add that image style configuration entity as dependency
      // of this widget.
      $dependencies[$style->getConfigDependencyKey()][] = $style->getConfigDependencyName();
    }
    return $dependencies;
  }

  /**
   * {@inheritdoc}
   */
  public function onDependencyRemoval(array $dependencies) {
    $changed = parent::onDependencyRemoval($dependencies);
    $style_id = $this->getSetting('preview_image_style');
    /** @var \Drupal\image\ImageStyleInterface $style */
    if ($style_id && $style = ImageStyle::load($style_id)) {
      if (!empty(
      $dependencies[$style->getConfigDependencyKey()][$style->getConfigDependencyName()]
      )) {
        /** @var \Drupal\image\ImageStyleStorageInterface $storage */
        $storage = $this->entityTypeManager->getStorage(
          $style->getEntityTypeId()
        );
        $replacement_id = $storage->getReplacementId($style_id);
        // If a valid replacement has been provided in the storage, replace the
        // preview image style with the replacement.
        if ($replacement_id && ImageStyle::load($replacement_id)) {
          $this->setSetting('preview_image_style', $replacement_id);
        }
        else {
          // If there's no replacement or the replacement is invalid,
          // disable the image preview.
          $this->setSetting('preview_image_style', '');
        }
        // Signal that the formatter plugin settings were updated.
        $changed = TRUE;
      }
    }
    return $changed;
  }

  /**
   * Overrides.
   *
   * \Drupal\file\Plugin\Field\FieldWidget\FileWidget::formMultipleElements().
   *
   * Special handling for draggable multiple widgets and 'add more' button.
   */
  protected function formMultipleElements(
    FieldItemListInterface $items,
    array &$form,
    FormStateInterface $form_state,
  ) {
    $elements = parent::formMultipleElements($items, $form, $form_state);

    $cardinality = $this->fieldDefinition->getFieldStorageDefinition()
      ->getCardinality();
    $file_upload_help = [
      '#theme' => 'file_upload_help',
      '#description' => '',
      '#upload_validators' => $elements[0]['#upload_validators'],
      '#cardinality' => $cardinality,
    ];
    if ($cardinality == 1) {
      // If there's only one field, return it as delta 0.
      if (empty($elements[0]['#default_value']['fids'])) {
        $file_upload_help['#description'] = $this->getFilteredDescription();
        $elements[0]['#description'] = $this->renderer->renderInIsolation($file_upload_help);
      }
    }
    else {
      $elements['#file_upload_description'] = $file_upload_help;
    }

    return $elements;
  }

}
